#include <u.h>
#include <libc.h>
#include <draw.h>
#include <thread.h>
#include <cursor.h>
#include <mouse.h>
#include <keyboard.h>
#include <frame.h>
#include <fcall.h>
#include <plumb.h>
#include "dat.h"
#include "fns.h"

enum {
	MShift   = 1<<0,
	MControl = 1<<1,
	MMeta    = 1<<2,
	MSuper   = 1<<5,  // Mod4 в Plan 9
	//
	Mmod4    = 1<<0,
	Mctl     = 1<<1,
	Mshift   = 1<<2,
};

enum {
	/*Spec    = 0xF800,*/
	Kmeta   = KF|0x15,
	Ksuper  = Spec|0x68,
	/*Kmod4=	Spec|0x68,*/
};

enum {
	MAXDESKTOPS = 9,
	STATUSHEIGHT = 20,
	ICONSIZE = 16,
	Maxtoken = 16,
	PADDING = 4,
};

typedef struct Window Window;
struct Window {
	int id;
	Rectangle rect;
	int desktop;
	Window* next;
};

typedef struct Desktop Desktop;
struct Desktop {
	Window* windows;
	int num;
};

typedef struct NotifyMsg NotifyMsg;
struct NotifyMsg {
	char text[256];
	uvlong timeout;
};

/* function declarations */
static void drawstatus(void);
static void drawworkspacebg(void);
static char *execstatus(void);
static int keyboardhandle(char *s);
static void keyboardthread(void* arg);
static int keyevent(char c, Rune r);
static void loadwallpaper(void);
static void switchdesk(int);
/*static void tilewindows(int);*/
static void winclose(void);
static Window *winnew(void);
static int winrun(char *name);
static int wsyswctl(int id, int mode);
static void wsysscan(int **ids, int *dsn);

/* variables */
static Keyboardctl *kbdctl;
static int mod = 0;
//static Mousectl *mousectl;
//
static Channel *notifychan;
static Desktop desks[MAXDESKTOPS];
static int currentdesk = 0;
static Image *statusbg = nil;
static Image *accentcol = nil;
static Image *greenbg = nil;
static Image *wallpaper = nil;
static char *wallpaperpath = nil;
static char *statuscmd = nil;
static Rectangle iconrects[MAXDESKTOPS];
/*static char notification[256];*/
/*static uvlong notification_ts;*/
/*static char *notifypath = "/dev/dwm_notify";*/
static int wsys;

/* function implementations */
/*static void
drawnotification(void)
{
	if(notification[0] == 0 || nsec() > notification_ts){
		notification[0] = 0;
		return;
	}

	Image* bg = allocimage(display, Rect(0,0,1,1), RGBA32, 1, 0x88888888);
	Rectangle r = Rect(10, STATUSHEIGHT + 10,
			   stringwidth(font, notification) + 20,
			   STATUSHEIGHT + 10 + font->height + 10);

	draw(screen, r, bg, nil, ZP);
	border(screen, r, 1, display->black, ZP);
	string(screen, addpt(r.min, Pt(10, 5)),
	       display->black, ZP, font, notification);
	freeimage(bg);
	}*/

static void
drawworkspacebg(void)
{
	if(desks[currentdesk].num > 0)
		return;

	Rectangle ws = Rect(0, STATUSHEIGHT,
			    screen->r.max.x, screen->r.max.y);

	if(wallpaper != nil){
		Image *scaled = allocimage(display, ws, RGB24, 0, DNofill);
		draw(scaled, scaled->r, wallpaper, nil, wallpaper->r.min);
		draw(screen, ws, scaled, nil, scaled->r.min);
		freeimage(scaled);
	} else {
		draw(screen, ws, greenbg, nil, ZP);
	}
}

static void
drawstatus(void)
{
	char *rightstatus = "nwm"; /* TODO: read from file */
	Point p = Pt(PADDING, 2);
	Font *f = font;

	// Создаём кастомные цвета
	Image *activefg = allocimage(display, Rect(0,0,1,1), RGB24, 1, 0xFFFFFFFF); // Белый
	Image *inactivefg = allocimage(display, Rect(0,0,1,1), RGB24, 1, 0x888888FF); // Серый
	Image *activebg = accentcol; // Оранжевый (уже инициализирован)
	Image *inactivebg = statusbg; // Чёрный (уже инициализирован)

	// Очищаем весь статусбар чёрным
	draw(screen, screen->r, inactivebg, nil, ZP);

	// Рисуем иконки рабочих столов
	for(int i = 0; i < MAXDESKTOPS; i++){
		Rectangle r = Rect(p.x, p.y, p.x+ICONSIZE, p.y+ICONSIZE);
		iconrects[i] = r;

		if(i == currentdesk){
			// Активный стол - оранжевый фон + белый текст
			draw(screen, r, activebg, nil, ZP);
			char num[2] = { '1'+i, 0 };
			Point tp = addpt(r.min, Pt(ICONSIZE/2-4, ICONSIZE/2-6));
			string(screen, tp, activefg, ZP, f, num);
		} else {
			// Неактивный стол - чёрный фон + серый текст
			draw(screen, r, inactivebg, nil, ZP);
			char num[2] = { '1'+i, 0 };
			Point tp = addpt(r.min, Pt(ICONSIZE/2-4, ICONSIZE/2-6));
			string(screen, tp, inactivefg, ZP, f, num);
		}

		p.x += ICONSIZE + PADDING;
	}

	// Правая часть статусбара (оранжевый фон + белый текст)
	int statuswidth = stringwidth(f, rightstatus) + 20;
	Rectangle statusr = Rect(
		screen->r.max.x - statuswidth,
		p.y,
		screen->r.max.x,
		p.y + ICONSIZE
		);

	draw(screen, statusr, activebg, nil, ZP);
	Point rp = Pt(statusr.min.x + 10, p.y + 2);
	string(screen, rp, activefg, ZP, f, rightstatus);

	// Фон рабочей области
	drawworkspacebg();

	// Уведомления
	/*drawnotification();*/

	flushimage(display, 1);

	// Освобождаем временные изображения
	freeimage(activefg);
	freeimage(inactivefg);
}

static void
keyboardthread(void *arg)
{
	char b[128];
	int n;
	int kbd_fd = open("/dev/kbd", OREAD);

	if (kbd_fd < 0)
		sysfatal("open kbd: %r");

	for(;;){
		if ((n = read(kbd_fd, b, sizeof(b)-1)) <= 0)
			break;
		b[n] = 0;
		if(keyboardhandle(b) != 0)
			break;
	}
	close(kbd_fd);
}

static int
keyboardhandle(char *s)
{
	char b[128], *p;
	int n, o;
	Rune r;

	o = 0;
	b[o++] = *s;
	if(*s == 'k' || *s == 'K'){
		mod = 0;
		if(utfrune(s+1, Kmod4) != nil)
			mod |= Mmod4;
		if(utfrune(s+1, Kctl) != nil)
			//mod |= Mctl;
			mod |= Mmod4;
		if(utfrune(s+1, Kshift) != nil)
			mod |= Mshift;
	}

	for(p = s+1; *p != 0; p += n){
		if((n = chartorune(&r, p)) == 1 && r == Runeerror){
			/* bail out */
			n = strlen(p);
			memmove(b+o, p, n);
			o += n;
			break;
		}

		if((mod & Mmod4) == 0 || keyevent(*s, r) != 0){
		/*if((mod & Mctl) == 0 || keyevent(*s, r) != 0){*/
			memmove(b+o, p, n);
			o += n;
		}
	}
	b[o++] = 0;
	/* TODO: Don't forward reserved keys */
	return (o > 1 && write(1, b, o) <= 0) ? -1 : 0;
}

static int
keyevent(char c, Rune r)
{
	if(c == 'c'){
		fprint(2, "evenr %d\n", r);
		if(r == '\n' && mod == Mmod4){
			winnew();
			return 0;
		}
		if(r == 'Q' && mod == (Mmod4|Mshift)){
			winclose();
			return 0;
		}
		if(r >= 17 && r <= 25){
			switchdesk(r - 17);
			return 0;
		}
	}
	/* skip redundant event */
	if(c == 'k' && (mod & Mshift) == 0)
		return 0;
	/* mod4 + shift + 1…0 yields a shifted value on 'c': workaround */
	if(c == 'k' && mod == (Mmod4|Mshift) && r >= '1' && r <= '9'){
		switchdesk(r - '1');
		return 0;
	}

	return -1;
}

static void
loadwallpaper(void)
{
	if(wallpaperpath == nil)
		return;

	int fd = open(wallpaperpath, OREAD);
	if(fd < 0){
		fprint(2, "failed to open wallpaper: %r\n");
		return;
	}

	wallpaper = readimage(display, fd, 0);
	close(fd);
}

/*static void
mousehandler(void* arg)
{
	Mouse m;
	while(recv(mousectl->c, &m)){
		if(m.buttons & 1){
			for(int i = 0; i < MAXDESKTOPS; i++){
				if(ptinrect(m.xy, iconrects[i])){
					switchdesk(i);
					break;
				}
			}
		}
	}
}*/

static void
switchdesk(int newdesk)
{
	int f;

	if(newdesk < 0 || newdesk >= MAXDESKTOPS)
		return;

	// Скрываем окна текущего стола
	for(Window *w = desks[currentdesk].windows; w; w = w->next){
		if((f = wsyswctl(w->id, OWRITE)) < 0)
			return;

		write(f, "hide", 4);
		close(f);
	}

	currentdesk = newdesk;

	// Показываем окна нового стола
	for(Window *w = desks[currentdesk].windows; w; w = w->next){
		if((f = wsyswctl(w->id, OWRITE)) < 0)
			return;

		write(f, "show", 4);
		close(f);
	}

	drawstatus();
}

static void
winclose(void)
{
	Window *w = nil, *prev = nil;

	// Ищем последнее окно в текущем рабочем столе
	for(Window *curr = desks[currentdesk].windows; curr != nil; curr = curr->next){
		prev = w;    // Сохраняем предыдущий элемент
		w = curr;    // Текущий элемент становится последним найденным
	}

	if(w == nil)
		return;      // Нет окон для закрытия

	// Посылаем команду закрытия
	int f;
	if((f = wsyswctl(w->id, OWRITE)) >= 0){
		write(f, "delete", 6);
		close(f);
	}

	// Удаляем из списка
	if(prev != nil)
		prev->next = nil;    // Отсоединяем последний элемент
	else
		desks[currentdesk].windows = nil;  // Список стал пустым

	free(w);
	desks[currentdesk].num--;
	drawstatus();    // Обновляем интерфейс
}

static int
winrun(char *name)
{
	int old_ids[50], new_ids[50];
	int old_count = 0, new_count;

	// Получаем текущий список ID
	wsysscan((int**)old_ids, &old_count);

	if(rfork(RFPROC|RFNOWAIT|RFNAMEG|RFENVG|RFCFDG|RFREND) == 0)
		execl("/bin/rc", "rc", name, nil);

	// Ждём появления нового ID
	for(;;){
		sleep(100);
		wsysscan((int**)new_ids, &new_count);

		// Ищем новый ID
		for(int i = 0; i < new_count; i++){
			int found = 0;
			for(int j = 0; j < old_count; j++){
				// Old ID
				if(new_ids[i] == old_ids[j]){
					found = 1;
					break;
				}
			}
			if(!found)
				return new_ids[i];
		}
	}
	return -1;
}

static Window*
winnew(void)
{
	Window *w = malloc(sizeof(Window));
	if(w == nil)
		sysfatal("malloc: %r");

	// Создаём окно и получаем его ID
	w->id = winrun("window");

	// Настройка параметров
	w->desktop = currentdesk;
	w->rect = Rect(0, STATUSHEIGHT, Dx(screen->r), Dy(screen->r)-STATUSHEIGHT);

	// Добавляем в список
	w->next = desks[currentdesk].windows;
	desks[currentdesk].windows = w;
	desks[currentdesk].num++;

	return w;
}

static int
wsyswctl(int id, int mode)
{
	char s[64];

	snprint(s, sizeof(s), "/dev/wsys/%d/wctl", id);

	return open(s, mode);
}

static void
wsysscan(int **ids, int *dsn)
{
	int i;
	Dir *ds, *d;
	*dsn = 0;

	seek(wsys, 0, 0);
	if((*dsn = dirreadall(wsys, &ds)) < 0)
		sysfatal("wsysscan: /dev/wsys: %r");

	for(i = 0, d = ds; i < *dsn; i++, d++)
		(*ids)[i] = atoi(d->name);

	free(ds);
}

static void
usage(void)
{
	fprint(2, "usage: %s\n", argv0);
	threadexitsall("usage");
}

void
threadmain(int argc, char* argv[])
{
	ARGBEGIN{
		case 'c':
			statuscmd = EARGF(usage());
			break;
			case 'w':
				wallpaperpath = EARGF(usage());
				break;
	}ARGEND

	if(initdraw(nil, nil, "dwm") < 0){
		fprint(2, "dwm: can't open display: %r\n");
		threadexitsall("initdraw");
	}

	if((wsys = open("/dev/wsys", OREAD)) < 0)
		sysfatal("open /dev/wsys: %r");

	statusbg = allocimage(display, Rect(0,0,1,1), RGB24, 1, 0x000000FF);
	accentcol = allocimage(display, Rect(0,0,1,1), RGB24, 1, 0xFFA500FF);
	greenbg = allocimage(display, Rect(0,0,1,1), RGB24, 1, 0x000000FF);

	loadwallpaper();
	drawstatus();

	/*keyboard.mods = 0;*/
	/*kbdctl = initkeyboard(nil);
	if(kbdctl == nil)
	sysfatal("initkeyboard failed");*/
	/*keyboard.keyc = chancreate(sizeof(Key), 0);
	  keyboard.mods = 0;*/
	proccreate(keyboardthread, nil, 8*1024);

	/*mousectl = initmouse(nil, screen);
	if(mousectl == nil)
		sysfatal("initmouse failed: %r");
		proccreate(mousehandler, nil, 4096);*/

	/*notifychan = chancreate(sizeof(NotifyMsg), 0);

	proccreate(notifyfs, nil, 4096);
	proccreate(notifyhandler, nil, 4096);*/

	for(;;){
		/*if(access(wallpaperpath, AREAD) == 0)
			loadwallpaper();
			drawstatus();*/
		sleep(5000);
	}
}
